;;; [c]2004 Andrzej Gumula, Katowice, Poland
;;; This routine creates curve text
;;; Accept control codes - degree, tolerance and diameter symbol
;;; Curve - for example a line, arc, circle, polyline, ellipse or spline, also entity of block (xref)
;;; Note - for knowledge, the routine always draw temporary circle at the beginning of the curve
;;; Please see CurveTextHelp.dwg file for help
;;; Files - CurveText.lsp, CurveText.dcl, CurveText.slb

(vl-load-com)

(defun c:CurveText ( / TxtPath TxtString TxtStart TxtPoint TxtChar TxtAngle TxtDir
                       CurvePoint LengthObj Flag Tmp Charwidth StartPt PickPoint
		       OldVars Olderr TmpObjectStart TmpObject CurveTxtHeight TxtStyle WhatNext
		       Id Fit x y)

(defun LispError (It)
 (if OldErr (setq *error* OldErr))
 (prompt (strcat "\nError: " (strcase It) "\n"))
 (if TmpObject (entdel TmpObject))
 (if TmpObjStart (entdel TmpObjStart))
 (if TxtPath (vla-highlight TxtPath 0))
 (vl-cmdf "_.undo" "_end")
 (mapcar 'setvar '("cmdecho" "highlight" "dimzin" "osmode") OldVars)
 (princ)
);end LispError

(defun CreateText (/ StringLength Item IndexChar TmpOffset TxtSet)
(while (/= TxtString (setq TxtString (vl-string-subst "" "%%u" TxtString))))
(while (/= TxtString (setq TxtString (vl-string-subst "" "%%U" TxtString))))
(while (/= TxtString (setq TxtString (vl-string-subst "" "%%o" TxtString))))
(while (/= TxtString (setq TxtString (vl-string-subst "" "%%O" TxtString))))
(if (minusp CurveTxtOffset) (setq CurveTxtOffset (- CurveTxtOffset CurveTxtHeight)))
(if (= TxtDir "1") (setq CurveTxtOffset (+ CurveTxtOffset CurveTxtHeight)))
(if (= TxtDir "1")
(setq Item (strlen TxtString) IndexChar -1)
(setq Item 1 IndexChar 1)
)
(cond
((= TxtStart "Middle") 
 (setq StringLength (* 0.5 (- LengthObj (TakeTmpWidth TxtString CurveTxtWidth))) 
 )
)
((= TxtStart "End")
 (setq StringLength (- LengthObj (TakeTmpWidth TxtString CurveTxtWidth))
 )
)
((and (= TxtStart "Pick") (= TxtOrient "End")) 
 (setq StringLength (vlax-curve-getDistAtPoint 
                     TxtPath 
                    (vlax-curve-getClosestPointTo TxtPath (trans PickPoint 1 0))
                   )
 )
)
((and (= TxtStart "Pick") (= TxtOrient "Begin")) 
 (setq StringLength (- (vlax-curve-getDistAtPoint 
                      TxtPath 
                      (vlax-curve-getClosestPointTo TxtPath (trans PickPoint 1 0))
                      )
                      (TakeTmpWidth TxtString CurveTxtWidth)
                   )

 )
)
(T (setq StringLength 0.0))
);end cond

(setq StringLength (+ (* 0.5 (TakeTmpWidth (substr TxtString Item 1) CurveTxtWidth))
                      StringLength
                   )
)
(repeat (strlen TxtString)
(cond
 ((and (= 1 IndexChar) (member (strcase (substr TxtString Item 3)) '("%%C" "%%D" "%%P" "%%%")))
  (setq TxtChar (substr TxtString Item 3))
  (setq Item (+ Item IndexChar))
  (setq Item (+ Item IndexChar))
 )
 ((and (= -1 IndexChar) (> Item 2 ) (member (strcase (substr TxtString (- Item 2) 3)) '("%%C" "%%D" "%%P" "%%%")))
  (setq TxtChar (substr TxtString (- Item 2) 3))
  (setq Item (+ Item IndexChar))
  (setq Item (+ Item IndexChar))
 )
 ((> Item 0) (setq TxtChar (substr TxtString Item 1)))
)
(Evaluation CharWidth)
(entmake (list '(0 . "TEXT") (cons 1 TxtChar) '(72 . 1) (cons 10 TxtPoint) (cons 7 (getvar "TEXTSTYLE"))
                (cons 11 TxtPoint) (cons 40 CurveTxtHeight) (cons 41 CurveTxtWidth)
                (cond ((= TxtDir "0") (cons 50 TxtAngle)) 
                      (T (cons 50 (- TxtAngle pi)))
                )
         )
)
(cond
 ((not TxtSet)  (setq TxtSet (ssadd (entlast))))
 (T (CorrectPos (entlast) (ssname TxtSet (1- (sslength TxtSet)))) (ssadd (entlast) TxtSet))
)
(setq Item (+ Item IndexChar))
(if (and (> Item 0) (<= Item (strlen TxtString)))
(setq StringLength 
 (+ StringLength 
  (setq CharWidth
   (if (minusp IndexChar) 
    (TakeTrueWidth (substr TxtString Item 1))
    (TakeTrueWidth TxtChar)
   )
  )
 )
)
);end if
);end repeat
(vl-cmdf "_.-group" "" (menucmd "M=$(edtime,$(getvar,date),D_MONTH_YY_HH_MM_SS)") "CurveTxt" TxtSet "")
);end CreateText


(defun XCopy (Elem / Matrix TMtx)
  (if (entmake (entget (car Elem)))
   (progn
   (setq Matrix (caddr Elem)
         TMtx (list (list (car (nth 0 Matrix)) (car (nth 1 Matrix)) (car (nth 2 Matrix)) (car (nth 3 matrix)))
		       (list (cadr (nth 0 Matrix)) (cadr (nth 1 Matrix)) (cadr (nth 2 Matrix)) (cadr (nth 3 matrix)))
		       (list (caddr (nth 0 Matrix)) (caddr (nth 1 Matrix)) (caddr (nth 2 Matrix)) (caddr (nth 3 matrix)))
		       '(0.0 0.0 0.0 1.0)))
   (if (vl-catch-all-error-p (vl-catch-all-apply 'vla-transformby (list (vlax-ename->vla-object (setq TmpObject (entlast))) (vlax-tmatrix TMtx))))
    (progn
     (entdel TmpObject)
     (setq TmpObject (prompt "\nSubentity with different X, Y scales. Please select another..."))
     (setq Check nil)
    )
   );end if
   )
  );end if
);end XCopy

(defun Dxf (Index IdElem)
 (cdr (assoc Index (entget IdElem)))
);end Dxf

 (defun ReactionButton (Reactor Pt)
  (setq Check T)
 );end ReactionButtom

(defun SelectObj (/ Object)
 (setq Check nil
       MouseReactor (VLR-Mouse-Reactor nil '((:VLR-beginRightClick . ReactionButton))))
 (while (not Check)
  (if (setq Object (nentsel "\nSelect entity (pick right button mouse to cancel command): "))
   (setq Check (member  (Dxf 0 (car Object)) '("POLYLINE" "LWPOLYLINE" "CIRCLE" "LINE" "ARC" "SPLINE" "ELLIPSE")))
  )
  (if (and Object (not Check)) 
   (prompt "\nPlease select line, arc, circle, polyline, ellipse or spline. ")
  )
  (if (> (length Object) 2)
   (Xcopy Object)
  );end if
 );end while
 (vlr-remove MouseReactor)
 (setq MouseReactor nil)
 (if (equal Check T)
  nil
  (progn
   (setq PickPoint (cadr Object))
   (if (> (length Object) 2)
    (vlax-ename->vla-object (entlast))
    (vlax-ename->vla-object (car Object))
   );end if
  )
 );end if
);end SelectObj

(defun TakeTmpWidth (Text Width / TxtBox) 
 (setq TxtBox 
  (textbox 
   (list (cons 1 Text) (cons 7 (getvar "TEXTSTYLE")) (cons 40 (if (numberp CurveTxtHeight) CurveTxtHeight (atof CurveTxtHeight))) 
         (cons 41 Width) '(50 . 0.0)
   )
  )
 )
 (- (caadr TxtBox) (caar TxtBox))
);end TmpWidth

(defun TakeTrueWidth (StrChar)
 (- (TakeTmpWidth (strcat "a" StrChar "a") CurveTxtWidth)
    (TakeTmpWidth "aa" CurveTxtWidth)
 )
);end TakeWidth

(defun CalculateWidth (Value)
 (cond
 ((= Value "1")
  (setq TxtStart "Begin")
  (set_tile "Begin" "1")
  (setq TxtOrient "End")
  (set_tile "ToEnd" "1")
  (mapcar 'mode_tile '("Pick" "Begin" "End" "Middle" "Width") '(1 1 1 1 1))
  
  (if (< (strlen TxtString) 3)
  (progn
   (mode_tile "accept" 1)
   (Error "In fit mode, user should enter at least 3 chars of text. " Key)
  )
  (progn
  (set_tile "Width" (setq CurveTxtWidth (R2S (/ LengthObj (TakeTmpWidth TxtString 1.0)))))
  (mode_tile "accept" 0)
  )
  );end if
   "1"
 )
 (T
  (mapcar 'mode_tile '("Pick" "Begin" "End" "Middle" "Width") '(0 0 0 0 0))
  "0"
 )
 );end cond
);end CalculateWidth

(defun TakeDeriv (Param)
 (angle '(0.0 0.0 0.0) 
         (vlax-curve-getFirstDeriv TxtPath Param)
 )
);end TakeDeriv

(defun GetStyles (/ StylesList Style)
(while (setq Style (tblnext "STYLE" (not Style)))
(if (equal 0 (logand 0 (cdr (assoc 70 Style))))
(setq StylesList (cons (cdr (assoc 2 Style)) StylesList))
);end if
);end while
(acad_strlsort StylesList)
);end GetStyles

(defun Evaluation (Add)
(cond
((and (>= StringLength 0) (>= LengthObj StringLength))
(setq TxtAngle (TakeDeriv (vlax-curve-getParamAtDist TxtPath StringLength))
      CurvePoint (vlax-curve-getPointAtDist TxtPath StringLength)
)
)
((minusp StringLength)
(setq TxtAngle (TakeDeriv (vlax-curve-getStartParam TxtPath))
      CurvePoint
      (cond
      ((not CurvePoint) (polar StartPt TxtAngle StringLength))
      (T (polar CurvePoint TxtAngle Add))
      )
)
)
(T 
(setq TxtAngle (TakeDeriv (vlax-curve-getEndParam TxtPath))
      CurvePoint (polar CurvePoint TxtAngle Add)
)
)
);end cond
(setq TxtPoint (polar CurvePoint (+ TxtAngle (/ pi 2)) CurveTxtOffset))
);end Evaluation

(defun CorrectPos (Current Prev / Old New CorrectDist)
(setq Old (entget Current)
      CorrectDist (- (distance (Dxf 11 Current) (Dxf 10 Current))
                     (distance (Dxf 11 Prev) (Dxf 10 Prev))
                  )
)
(if (= TxtDir "1") (setq CorrectDist (- CorrectDist)))
(setq  StringLength (+ StringLength CorrectDist))
(Evaluation CorrectDist)

(setq New (subst (cond ((= TxtDir "0") (cons 50 TxtAngle)) 
                       (T (cons 50 (- TxtAngle pi)))
                 )
                  (assoc 50 Old) Old
          )
)
(entmod (subst (cons 11 TxtPoint) (assoc 11 New) New))
);end CorrectPos

(defun R2S (Value)
(rtos Value (getvar "LUNITS") (getvar "LUPREC"))
);end R2S

(defun Error (Msg Key)
(set_tile "error" Msg)
(mode_tile Key 2)
(mode_tile Key 3)
nil
);end Error

(defun CheckValue (Value Key)
    (cond
     ((or (and (distof Value) (> (distof Value) 0.0)) (and (distof Value) (= Key "Offset")))
      (set_tile "error" "")
      (set_tile Key (R2S (distof Value)))
      (R2S (distof Value))
     )
     ((and (distof Value) (<= (distof Value) 0.0))
      (Error "Please enter a positive number. " Key)
     )
     (T (Error "Please enter a number. " Key))
    );end cond 
);end CheckValue

(defun CheckTxt (Value Key)
(if (= (setq TxtString (vl-string-trim " " Value)) "") (mode_tile "accept" 1) (mode_tile "accept" 0))
(if (= Fit "1") (CalculateWidth Fit))
(if (and (< (strlen TxtString) 3) (= Fit "1"))
  (progn
   (mode_tile "accept" 1)
   (Error "In fit mode, user should enter at least 3 chars of text. " Key)
  )
  (progn
   (mode_tile "accept" 0)
   (set_tile "error" "")
  )
);end if
(if (> (strlen TxtString) 2)
  (progn 
   (mode_tile "Fit" 0)
   ;;(set_tile "Fit" "0")
  )
  (progn
    (mode_tile "Fit" 1)
  )
);end if
);CheckTxt

(defun ShowImage ()
 (if (= TxtStart "Pick")
   (mapcar 'mode_tile '("ToEnd" "ToBegin") '(0 0))
   (mapcar 'mode_tile '("ToEnd" "ToBegin") '(1 1))
 );end if
 (start_image "Image")
 (fill_image 0 0 x y -15)
 (cond
  ((and (= TxtDir "0") (>= (atof CurveTxtOffset) 0.0) (= Fit "1"))
   (slide_image 0 0 x y "CurveText(13)")
  )
  ((and (= TxtDir "0") (< (atof CurveTxtOffset) 0.0) (= Fit "1"))
   (slide_image 0 0 x y "CurveText(14)")
  )
  ((and (= TxtDir "1") (>= (atof CurveTxtOffset) 0.0) (= Fit "1"))
   (slide_image 0 0 x y "CurveText(15)")
  )
  ((and (= TxtDir "1") (< (atof CurveTxtOffset) 0.0) (= Fit "1"))
   (slide_image 0 0 x y "CurveText(16)")
  )
  ((and (= TxtDir "0") (>= (atof CurveTxtOffset) 0.0) (= TxtStart "Middle"))
   (slide_image 0 0 x y "CurveText(1)")
  )
  ((and (= TxtDir "0") (< (atof CurveTxtOffset) 0.0) (= TxtStart "Middle"))
   (slide_image 0 0 x y "CurveText(2)")
  )
  ((and (= TxtDir "1") (>= (atof CurveTxtOffset) 0.0) (= TxtStart "Middle"))
   (slide_image 0 0 x y "CurveText(3)")
  )
  ((and (= TxtDir "1") (< (atof CurveTxtOffset) 0.0) (= TxtStart "Middle"))
   (slide_image 0 0 x y "CurveText(4)")
  )
  ((and (= TxtDir "0") (>= (atof CurveTxtOffset) 0.0) (= TxtStart "Begin"))
   (slide_image 0 0 x y "CurveText(5)")
  )
  ((and (= TxtDir "0") (< (atof CurveTxtOffset) 0.0) (= TxtStart "Begin"))
   (slide_image 0 0 x y "CurveText(6)")
  )
  ((and (= TxtDir "1") (>= (atof CurveTxtOffset) 0.0) (= TxtStart "Begin"))
   (slide_image 0 0 x y "CurveText(7)")
  )
  ((and (= TxtDir "1") (< (atof CurveTxtOffset) 0.0) (= TxtStart "Begin"))
   (slide_image 0 0 x y "CurveText(8)")
  )
  ((and (= TxtDir "0") (>= (atof CurveTxtOffset) 0.0) (= TxtStart "End"))
   (slide_image 0 0 x y "CurveText(9)")
  )
  ((and (= TxtDir "0") (< (atof CurveTxtOffset) 0.0) (= TxtStart "End"))
   (slide_image 0 0 x y "CurveText(10)")
  )
  ((and (= TxtDir "1") (>= (atof CurveTxtOffset) 0.0) (= TxtStart "End"))
   (slide_image 0 0 x y "CurveText(11)")
  )
  ((and (= TxtDir "1") (< (atof CurveTxtOffset) 0.0) (= TxtStart "End"))
   (slide_image 0 0 x y "CurveText(12)")
  )
  ((and (= TxtDir "0") (= TxtOrient "End") (>= (atof CurveTxtOffset) 0.0) (= TxtStart "Pick"))
   (slide_image 0 0 x y "CurveText(17)")
  )
  ((and (= TxtDir "0") (= TxtOrient "End") (< (atof CurveTxtOffset) 0.0) (= TxtStart "Pick"))
   (slide_image 0 0 x y "CurveText(18)")
  )
  ((and (= TxtDir "1") (= TxtOrient "End") (>= (atof CurveTxtOffset) 0.0) (= TxtStart "Pick"))
   (slide_image 0 0 x y "CurveText(19)")
  )
  ((and (= TxtDir "1") (= TxtOrient "End") (< (atof CurveTxtOffset) 0.0) (= TxtStart "Pick"))
   (slide_image 0 0 x y "CurveText(20)")
  )

  ((and (= TxtDir "0") (= TxtOrient "Begin") (>= (atof CurveTxtOffset) 0.0) (= TxtStart "Pick"))
   (slide_image 0 0 x y "CurveText(21)")
  )
  ((and (= TxtDir "0") (= TxtOrient "Begin") (< (atof CurveTxtOffset) 0.0) (= TxtStart "Pick"))
   (slide_image 0 0 x y "CurveText(22)")
  )
  ((and (= TxtDir "1") (= TxtOrient "Begin") (>= (atof CurveTxtOffset) 0.0) (= TxtStart "Pick"))
   (slide_image 0 0 x y "CurveText(23)")
  )
  ((and (= TxtDir "1") (= TxtOrient "Begin") (< (atof CurveTxtOffset) 0.0) (= TxtStart "Pick"))
   (slide_image 0 0 x y "CurveText(24)")
  )
 );end cond
 (end_image)
);end ShowImage

(setq OldErr *error* *error* LispError)
(setq OldVars (mapcar 'getvar '("cmdecho" "highlight" "dimzin" "osmode")))
(mapcar 'setvar '("cmdecho" "highlight" "dimzin" "osmode") '(0 1 0 0))
(vl-cmdf "_.undo" "_be")

(cond
((setq TxtPath (SelectObj))
 (setq StartPt (vlax-curve-getStartPoint TxtPath)
       LengthObj (vlax-curve-getDistAtParam TxtPath (vlax-curve-getEndParam TxtPath))
 )
 (entmake (list '(0 . "CIRCLE") (cons 10 StartPt)
                 (cons 40 (/ (getvar "VIEWSIZE") 100.0)) (cons 62 (vla-get-color TxtPath))
          )
 )
 (setq TmpObjStart (entlast))
 (vla-highlight TxtPath 1)
 (princ)
 (setq TxtString "")

 (setq CurveTxtHeight (R2S (getvar "TEXTSIZE")))
 (if (or (not CurveTxtWidth) (not (numberp CurveTxtWidth)) (<= CurveTxtWidth 0.0))
  (setq CurveTxtWidth (R2S 1.0))
  (setq CurveTxtWidth (R2S CurveTxtWidth))
 );end if
 (if (or (not CurveTxtOffset) (not (numberp CurveTxtOffset)))
  (setq CurveTxtOffset (R2S 0.0))
  (setq CurveTxtOffset (R2S CurveTxtOffset))
 );end if
 (setq TxtStyle (getvar "TEXTSTYLE")
       TxtStart "Middle"
       TxtDir "0"
       TxtOrient "End")
 (setq Id (load_dialog "CurveText"))
 (if (not (new_dialog "CURVETEXT" Id))
  (progn
    (alert "Cannot open CurveText.dcl file. ")
    (exit)
  )
 );end if
 (if (not (findfile "CurveText.slb"))
   (alert "Cannot find CurveText.slb file.\nThis file is necessary to correct display of dialog window.")
 );end if
 (mapcar 'set_tile '("error" "Text" "Height" "Width" "Offset" "TxtStyle")
                    (list "" TxtString CurveTxtHeight CurveTxtWidth CurveTxtOffset (strcat "Current style text: " TxtStyle)))
 (if (= (vl-string-trim " " TxtString) "") (mode_tile "accept" 1) (mode_tile "accept" 0))
 (setq x (dimx_tile "Image")
       y (dimy_tile "Image"))
 (start_image "Image")
 (fill_image 0 0 x y -15)
 (slide_image 0 0 x y "CurveText(1)")
 (end_image)
 (start_list "Styles")
 (mapcar 'add_list (GetStyles))
 (end_list)

 (action_tile "Text"  "(CheckTxt (get_tile $key) $key)")
 (action_tile "Offset"  "(setq CurveTxtOffset (CheckValue (get_tile $key) $key)) (ShowImage)")
 (action_tile "Height"  "(setq CurveTxtHeight (CheckValue (get_tile $key) $key)) (CalculateWidth Fit)")
 (action_tile "Width"  "(setq CurveTxtWidth (CheckValue (get_tile $key) $key))")
 (action_tile "Styles" "(set_tile \"TxtStyle\" (strcat \"Current style text: \" (setq TxtStyle (nth (atoi $value) (GetStyles)))))")
 (action_tile "Pick" "(setq TxtStart \"Pick\") (ShowImage)")
 (action_tile "Begin" "(setq TxtStart \"Begin\") (ShowImage)")
 (action_tile "End" "(setq TxtStart \"End\") (ShowImage)")
 (action_tile "Middle" "(setq TxtStart \"Middle\") (ShowImage)")
 (action_tile "ToEnd" "(setq TxtOrient \"End\") (ShowImage)")
 (action_tile "ToBegin" "(setq TxtOrient \"Begin\") (ShowImage)")
 (action_tile "Reverse" "(setq TxtDir $value) (ShowImage)")
 (action_tile "Fit" "(setq Fit (CalculateWidth $value)) (ShowImage)")
 (action_tile "accept" "(done_dialog 1)")
 (action_tile "cancel" "(done_dialog 0)")
 (setq WhatNext (start_dialog))

 (cond
 ((= WhatNext 1)	
 (setvar "TEXTSTYLE" TxtStyle)
 (setvar "TEXTSIZE" (setq CurveTxtHeight (atof CurveTxtHeight)))
 (setq CurveTxtWidth (atof CurveTxtWidth)) 
 (setq CurveTxtOffset (atof CurveTxtOffset))
 (done_dialog 1)
 (unload_dialog Id)
 (CreateText)
 )
 ((= WhatNext 0)
 (prompt "\nCommand canceled. ")
 ) 
 );end cond
 (vla-highlight TxtPath 0)
 (if TmpObjStart (entdel TmpObjStart))
 (if TmpObject (entdel TmpObject))
 )
);end cond
(setq *error* OldErr)
(vl-cmdf "_.undo" "_end")
(mapcar 'setvar '("cmdecho" "highlight" "dimzin" "osmode") OldVars)
(princ)
);end file

(prompt "\nLoaded new command CurveText. ")
(prompt "\n[c]2004 Andrzej Gumula. ")
(princ)

